package sketch;

import java.io.File;
import java.util.ArrayList;
import java.util.Arrays;
import java.util.Locale;
import java.util.Random;

import fileIO.TextStreamWriter;
import shared.Shared;
import shared.Tools;
import tax.PrintTaxonomy;
import tax.TaxNode;
import tax.TaxTree;

public class SketchObject {
	
	/*--------------------------------------------------------------*/
	/*----------------           Parsing            ----------------*/
	/*--------------------------------------------------------------*/
	
	public static boolean parseSketchFlags(String arg, String a, String b){
		
		if(parseCoding(arg, a, b)){
			//
		}
		
		else if(a.equalsIgnoreCase("k")){
			k=Integer.parseInt(b);
		}else if(a.equalsIgnoreCase("rcomp")){
			rcomp=Tools.parseBoolean(b);
		}else if(a.equalsIgnoreCase("minfakeid")){
			minFakeID=Integer.parseInt(b);
		}else if(a.equalsIgnoreCase("hashNames")){
			hashNames=Tools.parseBoolean(b);
		}else if(a.equalsIgnoreCase("skipNonCanonical")){
			skipNonCanonical=Tools.parseBoolean(b);
		}else if(a.equalsIgnoreCase("minfakeid")){
			minFakeID=Integer.parseInt(b);
		}else if(a.equalsIgnoreCase("blacklist")){
			blacklist=b;
			if(b!=null){
				if("nt".equalsIgnoreCase(b) && !new File(b).exists()){
					blacklist=Blacklist.ntBlacklist;
				}else if(("silva".equalsIgnoreCase(b) || "ribo".equalsIgnoreCase(b)) && !new File(b).exists()){
					blacklist=Blacklist.silvaBlacklist;
				}else if("refseq".equalsIgnoreCase(b) && !new File(b).exists()){
					blacklist=Blacklist.refseqBlacklist;
				}else if("img".equalsIgnoreCase(b) && !new File(b).exists()){
					blacklist=Blacklist.imgBlacklist;
				}
			}
		}else if(a.equalsIgnoreCase("whitelist") || a.equalsIgnoreCase("usewhitelist")){
			useWhitelist=Tools.parseBoolean(b);
		}
		
		else if(a.equalsIgnoreCase("reads")){
			maxReads=Long.parseLong(b);
		}else if(a.equalsIgnoreCase("samplerate")){
			samplerate=Float.parseFloat(b);
		}else if(a.equalsIgnoreCase("sampleseed")){
			sampleseed=Long.parseLong(b);
		}
		
		else if(a.equalsIgnoreCase("size") || a.equalsIgnoreCase("sketchsize")){
			if("auto".equalsIgnoreCase(b)){
				AUTOSIZE=true;
			}else{
				AUTOSIZE=false;
				size=Integer.parseInt(b);
			}
		}else if(a.equalsIgnoreCase("maxfraction") || a.equalsIgnoreCase("maxgenomefraction") || a.equalsIgnoreCase("mgf")){
			maxGenomeFraction=Float.parseFloat(b);
		}else if(a.equalsIgnoreCase("smallsketchsize")){
			smallSketchSize=Integer.parseInt(b);
		}else if(a.equalsIgnoreCase("minSketchSize")){
			minSketchSize=Tools.max(1, Integer.parseInt(b));
		}else if(a.equalsIgnoreCase("autosize")){
			AUTOSIZE=Tools.parseBoolean(b);
		}else if(a.equalsIgnoreCase("autosizefactor")){
			AUTOSIZE_FACTOR=Float.parseFloat(b);
		}else if(a.equalsIgnoreCase("maxGenomeFractionSmall")){
			maxGenomeFractionSmall=Float.parseFloat(b);
		}else if(a.equalsIgnoreCase("keyFraction")){
			double d=Double.parseDouble(b);
			setKeyFraction(d);
		}
		
//		else if(a.equalsIgnoreCase("minHashValue")){
//			minHashValue=Tools.max(1, Long.parseLong(b));
//		}
		
		else{
			return false;
		}
		return true;
	}
	
	private static boolean parseCoding(String arg, String a, String b){
		if(a.equalsIgnoreCase("delta")){
			delta=Tools.parseBoolean(b);
		}else if(a.equalsIgnoreCase("a33") || a.equalsIgnoreCase("a48")){
			boolean x=Tools.parseBoolean(b);
			if(x){CODING=A48;}
			else if(CODING==A48){CODING=HEX;}
		}else if(a.equalsIgnoreCase("hex")){
			boolean x=Tools.parseBoolean(b);
			if(x){CODING=HEX;}
			else if(CODING==HEX){CODING=A48;}
		}else if(a.equalsIgnoreCase("raw")){
			boolean x=Tools.parseBoolean(b);
			if(x){CODING=RAW;}
			else if(CODING==RAW){CODING=A48;}
		}else{
			return false;
		}
		return true;
	}
	
	static int parseMode(String arg, String a, String b){
		int mode_=-1;
		if(a.equalsIgnoreCase("mode")){
			if(b!=null && Character.isDigit(b.charAt(0))){
				mode_=Integer.parseInt(b);
			}else if(b.equalsIgnoreCase("single") || b.equalsIgnoreCase("onesketch")){
				mode_=ONE_SKETCH;
			}else if(b.equalsIgnoreCase("taxa") || b.equalsIgnoreCase("pertaxa")){
				mode_=PER_TAXA;
			}else if(b.equalsIgnoreCase("sequence") || b.equalsIgnoreCase("persequence")){
				mode_=PER_SEQUENCE;
			}else if(b.equalsIgnoreCase("img")){
				mode_=IMG;
			}else if(Character.isDigit(b.charAt(0))){
				mode_=Integer.parseInt(b);
			}
		}else if(a.equalsIgnoreCase("single") || a.equalsIgnoreCase("onesketch")){
			mode_=ONE_SKETCH;
		}else if(a.equalsIgnoreCase("taxa") || a.equalsIgnoreCase("pertaxa")){
			mode_=PER_TAXA;
		}else if(a.equalsIgnoreCase("sequence") || a.equalsIgnoreCase("persequence")){
			mode_=PER_SEQUENCE;
		}
		return mode_;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------        Initialization        ----------------*/
	/*--------------------------------------------------------------*/
	
	static synchronized void setTaxtree(String taxTreeFile){
		if(taxTreeFile==null){
			taxtree=null;
			return;
		}
		if(treefile!=null){
			assert(!treefile.equals(taxTreeFile));
			if(treefile.equals(taxTreeFile)){return;}
			treefile=taxTreeFile;
		}
		taxtree=TaxTree.loadTaxTree(taxTreeFile, System.err, hashNames);
	}
	
	public static void postParse(){
		if(amino){
			k=Tools.min(k, 12);
			rcomp=false;
		}
		if(blacklist!=null){Blacklist.addFiles(blacklist);}
	}
	
	/*--------------------------------------------------------------*/
	/*----------------          Formatting          ----------------*/
	/*--------------------------------------------------------------*/
	
	static void formatComparisonColumnwise(Comparison c, StringBuilder sb, boolean printTax, boolean printOriginalName){
			sb.append(String.format(Locale.ROOT, "%.2f%%\t%.2f%%\t%d\t%d\t%d\t%d\t%d\t",
					100*c.idWeighted(), 100*c.idMin(), c.hits, c.minIndex, c.taxID, c.genomeSizeBases, c.genomeSizeKmers));
			if(taxtree==null || !printTax){
				sb.append(c.taxName==null ? '.' : c.taxName);
			}else{
				if(c.taxID>=0 && c.taxID<minFakeID){
					TaxNode tn=taxtree.getNode(c.taxID);
					if(tn!=null){
						sb.append(taxtree.toSemicolon(tn, skipNonCanonical));
					}
				}
			}
			if(printOriginalName){
				sb.append('\t');
				sb.append(c.name0==null ? "." : c.name0);
			}
			sb.append('\n');
	}
	
	static String header(boolean printOriginalName){
		String s="WKID\tKID\tmatches\tcompared\ttaxID\tgSize\tgKmers\ttaxName"+(printOriginalName ? "\tseqName" : "");
		return s;
	}
	
	static void formatComparison(Comparison c, int format, StringBuilder sb, boolean printTax, boolean printOriginalName){
		if(format==2){
			formatComparisonColumnwise(c, sb, printTax, printOriginalName);
			return;
		}
		if(format==0){
			sb.append(String.format(Locale.ROOT, "WKID %.2f%%\tKID %.2f%%\tmatches %d\tcompared %d",
					100*c.idWeighted(), 100*c.idMin(), c.hits, c.minIndex)+"\ttaxID "+c.taxID+"\tgKmers "+c.genomeSizeKmers+"\t"+
					(c.taxName==null ? "." : c.taxName)+
					(printOriginalName ? "\t"+(c.name0==null ? "." : c.name0) : "")+"\n");
			if(printTax && taxtree!=null){
				if(c.taxID>=0 && c.taxID<minFakeID){
					TaxNode tn=taxtree.getNode(c.taxID);
					if(tn!=null){
						PrintTaxonomy.printTaxonomy(tn, sb, taxtree, TaxTree.DOMAIN, skipNonCanonical);
					}
				}
				sb.append('\n');
			}
		}else{
			ArrayList<TaxNode> tnl=new ArrayList<TaxNode>();
			if(taxtree!=null && c.taxID>=0 && c.taxID<minFakeID){
				TaxNode tn=taxtree.getNode(c.taxID);
				while(tn!=null && tn.pid!=tn.id && tn.level<=TaxTree.DOMAIN){
					tnl.add(tn);
					tn=taxtree.getNode(tn.pid);
				}
			}
			
			sb.append(String.format(Locale.ROOT, "WKID %.2f%%\tKID %.2f%%\tmatches %d\tcompared %d\t",
					100*c.idWeighted(), 100*c.idMin(), c.hits, c.minIndex));
			sb.append("\ttaxID ").append(c.taxID).append('\t');
			sb.append(c.taxName).append('\t');
			if(printOriginalName){sb.append(c.name0).append('\t');}
			
			for(int i=tnl.size()-1; i>=0; i--){
				TaxNode tn=tnl.get(i);
				sb.append(tn.name);
				if(i>0){sb.append(';');}
			}
			sb.append('\n');
			
			tnl.clear();
		}
	}

	
	/*--------------------------------------------------------------*/
	/*----------------         Print Methods        ----------------*/
	/*--------------------------------------------------------------*/
	
	void writeResults(ArrayList<Comparison> al, Sketch s, int format, boolean printTax, boolean printOriginalName, TextStreamWriter tsw){
		tsw.print("\nResults for "+s.name()+":\n\n");
		
		StringBuilder sb=new StringBuilder();
		if(format==2){sb.append(header(printOriginalName)).append('\n');}
		for(Comparison c : al){	
			formatComparison(c, format, sb, printTax, printOriginalName);
			tsw.print(sb.toString());
			sb.setLength(0);
		}
	}
	
	StringBuilder toText(ArrayList<Comparison> al, Sketch s, int format, boolean printTax, boolean printOriginalName){
		StringBuilder sb=new StringBuilder();
		
		sb.append("\nResults for "+s.name()+":\n\n");

		if(format==2){sb.append(header(printOriginalName)).append('\n');}
		for(Comparison c : al){	
			formatComparison(c, format, sb, printTax, printOriginalName);
		}
		return sb;
	}
	
	/*--------------------------------------------------------------*/
	/*----------------           Hashing            ----------------*/
	/*--------------------------------------------------------------*/
	
	public static long[][] makeCodes(int symbols, int modes, long seed, boolean positive){
		Random randy=(seed>=0 ? new Random(seed) : new Random());
		long mask=positive ? Long.MAX_VALUE : -1L;
		long[][] r=new long[symbols][modes];
		for(int i=0; i<symbols; i++){
			for(int j=0; j<modes; j++){
				r[i][j]=randy.nextLong()&mask;
			}
		}
		return r;
	}
	
	
	/**
	 * Generates a hash code from a kmer.
	 * @param kmer Kmer to hash
	 * @return Hash code
	 */
	public static final long hash(long kmer){
		long code=kmer;
		for(int i=0; i<8; i++){
			int x=(int)(kmer&0xFF);
			kmer>>=8;
			code^=codes[i][x];
//			code^=Long.rotateRight(codes[i][x], (int)(code&3)); //Might be better.
		}
		return code;
	}
	
	private static final long[][] codes=makeCodes(8, 256, 12345, false);

	
	/*--------------------------------------------------------------*/
	/*----------------      Convenience Methods     ----------------*/
	/*--------------------------------------------------------------*/

	public static void kmerArrayToSketchArray(long[] kmers){
		for(int i=0; i<kmers.length; i++){
			long kmer=kmers[i];
			long hash=SketchTool.hash(kmer);
			assert(hash>=minHashValue);
			hash=Long.MAX_VALUE-hash;
			kmers[i]=hash;
		}
		Shared.sort(kmers);
	}
	
	public static final int toSketchSize(long genomeSizeBases, long genomeSizeKmers, int maxSketchSize){
		if(genomeSizeKmers>0){
			return toSketchSizeKmers(genomeSizeKmers, maxSketchSize);
		}
		return toSketchSizeBases(genomeSizeBases, maxSketchSize);
	}
	
	private static final int toSketchSizeBases(long genomeSizeBases, int maxSketchSize) {
		return toSketchSizeKmers(Tools.max(0, genomeSizeBases-k+1), maxSketchSize);
	}
	
	private static final int toSketchSizeKmers(long genomeSizeKmers, int maxSketchSize) {
		if(AUTOSIZE){
			float linear1=Tools.min(smallSketchSize+0.5f*(float)Math.sqrt(genomeSizeKmers), 
					maxGenomeFractionSmall*genomeSizeKmers-10);
			float linear2=genomeSizeKmers*maxGenomeFraction;
			float poly=0+1f*(float)Math.sqrt(genomeSizeKmers)+90f*(float)Math.pow(genomeSizeKmers, 0.3);
			float log=Tools.max(1000, -4000+3500*(float)Math.log(Tools.max(1, genomeSizeKmers)));
			float min=Tools.min(Tools.max(linear1, linear2), poly, log);
			assert(min>=Integer.MIN_VALUE && min<=Integer.MAX_VALUE) : min;
			
//			System.err.println(genomeSizeKmers+" -> "+linear1+", "+linear2+", "+poly+", "+log);
			
			return (int)Tools.min(genomeSizeKmers*keyFraction2, min*AUTOSIZE_FACTOR);
//			return (int)Tools.max(1, min*AUTOSIZE_FACTOR);
		}else{
			return Tools.min((int)(2+maxGenomeFraction*genomeSizeKmers), maxSketchSize);
		}
	}
	
	/*--------------------------------------------------------------*/
	/*----------------          Constants           ----------------*/
	/*--------------------------------------------------------------*/
	
	public static final int RAW=0, HEX=1, A48=2;
	public static final char[] codingArray={'R', 'H', 'A'};
	
	static final byte[] hexTable=new byte[128];
	static {
		Arrays.fill(hexTable, (byte)-1);
		for(int i='0'; i<='9'; i++){
			hexTable[i]=(byte)(i-'0');
		}
		for(int i='A'; i<='F'; i++){
			hexTable[i]=hexTable[i+'a'-'A']=(byte)(i-'A'+10);
		}
		hexTable['x']=hexTable['X']=hexTable['-']=hexTable['+']=0;
	}
	
	public static final int ONE_SKETCH=1, PER_SEQUENCE=2, PER_TAXA=3, IMG=4;
	
	/*--------------------------------------------------------------*/
	/*----------------        Static Fields         ----------------*/
	/*--------------------------------------------------------------*/
	
	public static DisplayParams defaultParams=new DisplayParams();
	
	public static int k=31;
	public static int CODING=A48;
	public static boolean delta=true;
	public static boolean rcomp=true;
	public static boolean amino=false;

	public static int minSketchSize=3;
	public static int size=10000;
	public static float maxGenomeFraction=0.01f;
	public static float maxGenomeFractionSmall=0.10f;
	public static int smallSketchSize=150;
	public static boolean AUTOSIZE=true;
	public static float AUTOSIZE_FACTOR=1;

	private static double keyFraction=0.2;
	private static double keyFraction2=keyFraction*1.2;
	public static long minHashValue=setMinHashValue();
	public static double keyFraction(){return keyFraction;}
	public static void setKeyFraction(double d){
		assert(d>0);
		keyFraction=Tools.mid(0, d, 0.5);
		keyFraction2=Tools.mid(0, keyFraction*1.2, 0.5);
	}
	public static long setMinHashValue(){
		double mult=1-2*keyFraction;
		minHashValue=(long)(mult*Long.MAX_VALUE);
		assert(minHashValue>=0 && minHashValue<Long.MAX_VALUE);
		return minHashValue;
	}
	
	static int minFakeID=1900000000;
	static boolean hashNames=false;
	static boolean skipNonCanonical=true;
	
	static long maxReads=-1;
	static float samplerate=1.0f;
	static long sampleseed=-1L;
	
	public static TaxTree taxtree=null;
	private static String treefile=null;
	static String blacklist=null;
	static boolean useWhitelist=false;
	
}
